"""
This module contains API models.
"""

# isort: dont-add-imports

from .artist import *
from .backgrounds import *
from .base import *
from .beatmap import *
from .changelog import *
from .chat import *
from .comment import *
from .common import *
from .event import *
from .files import *
from .forum import *
from .gamemode import *
from .kudosu import *
from .lazer import *
from .mods import *
from .multiplayer import *
from .news import *
from .oauthtoken import *
from .performance import *
from .rankings import *
from .scopes import *
from .score import *
from .search import *
from .spotlight import *
from .user import *
from .wiki import *
